from sec_parsers.string_detector_groups import HeaderStringDetectorGroup, SEC_10K_StringDetectorGroup, SEC_8K_StringDetectorGroup, SEC_S1_StringDetectorGroup
from sec_parsers.xml_helper import get_text, get_all_text,\
        set_background_color, remove_background_color, open_tree, is_middle_element
from sec_parsers.cleaning import clean_title
from sec_parsers.visualization_helper import headers_colors_list
from lxml import etree

from sec_parsers.element_detector_groups import HeaderElementDetectorGroup, SEC_10K_ElementGroup, SEC_8K_ElementGroup

# TODO: refactor to reduce code duplication / annoying stuff like detectors all over the place

# TODO, probably add autosort later
class HTMLParser:
    def __init__(self, **kwargs):
        self.element_detector_group = HeaderElementDetectorGroup() # parses elements
        self.string_detector_group = HeaderStringDetectorGroup() # parses strings
        self.color_dict = {'remove;': '#f2f2f2','skip;':'#FAFAD2'} # If you want certain parsing types to have certain colors

        self._init(**kwargs)
        self.all_detectors = self.element_detector_group.element_detectors + self.string_detector_group.string_detectors
        self.sort_detectors()

    def _init(self, **kwargs):
        # This method is meant to be overridden by subclasses
        pass

    def sort_detectors(self):
        """sorts detector by parsing rule. Return first"""

        # Separate detectors into 'return' and 'continue' groups
        return_detectors = [detector for detector in self.all_detectors if detector.parsing_rule == "return"]
        continue_detectors = [detector for detector in self.all_detectors if detector.parsing_rule == "continue"]

        # Combine the groups, with 'return' detectors first
        self.all_detectors = return_detectors + continue_detectors


    def detect_style_from_string(self,string,rule_list=[]):
        string_detectors = self.string_detector_group.string_detectors
        # subset
        if rule_list: # WIP
            string_detectors = [detector for detector in string_detectors if detector.parsing_rule in rule_list]
        
        result = ''
        for detector in string_detectors:
            string_style = detector.detect(string)
            if string_style != '':
                parsing_rule = detector.parsing_rule
                if parsing_rule == 'return':
                    return string_style, parsing_rule
                else:
                    result += string_style
        
        return (result,'continue')
    
    
    def detect_style_from_element(self,element,rule_list=[]):
        element_detectors = self.element_detector_group.element_detectors
        # subset
        if rule_list: # WIP
            element_detectors = [detector for detector in element_detectors if detector.parsing_rule in rule_list]
        
        result = ''
        for detector in element_detectors:
            element_style = detector.detect(element)
            if element_style != '':
                parsing_rule = detector.parsing_rule
                if parsing_rule == 'return':
                    return element_style, parsing_rule
                else:
                    result += element_style
        
        return (result,'continue')
    
    # possible performance increases - move which detector triggers first
    def iterative_parse(self,html,add_parsing_id=False):
        count = 0

        flag = True
        orig_elem = None
        element_style = ''
        string_style = ''
        parser = self
        # add xml construction later?
        for event, elem in etree.iterwalk(html, events=('start', 'end')): # fast, .7s
            if event == 'start':
                if add_parsing_id:
                    elem.attrib['parsing_id'] = str(count)
                    count += 1  
                    
                if flag:
                    # refactor to seperate function
                    result, parsing_rule = parser.detect_style_from_element(elem)
                    if parsing_rule == 'return':
                        element_style = result
                        orig_elem = elem
                        flag = False
                        continue
                    elif result != '':
                        if orig_elem is None:
                            orig_elem = elem

                        element_style += result

                    if string_style == '':
                        string = get_all_text(elem)
                        result, parsing_rule = parser.detect_style_from_string(string)
                        if parsing_rule == 'return':
                            string_style = result
                            orig_elem = elem
                            flag = False
                            continue
                        elif result != '':
                            if orig_elem is None:
                                orig_elem = elem

                            string_style += result

                    # code to check if elem is in middle 
                    if (element_style + string_style) != '':
    
                        if is_middle_element(orig_elem):
                            element_style =''
                            string_style = ''
                            orig_elem = None

                else:
                    pass # iterate through file without parsing

            elif event == 'end':
                if orig_elem is not None:
                    if elem == orig_elem:
                        parsing_string = element_style + string_style
                        orig_elem.attrib['parsing_string'] = parsing_string

                        flag = True
                        orig_elem = None
                        string_style = ''
                        element_style = ''

    # Works
    def clean_parse(self,html):
        """set ignore items etc"""
        parsed_elements_to_pop = html.xpath("//*[@parsing_string='']") # should be able to delete this now
        for element in parsed_elements_to_pop:
            element.attrib.pop('parsing_string', None)

        parsed_elements = html.xpath('//*[@parsing_string]')

        # figure out how to manage remove, skip, and header strigns here
        remove_strings = [item.style for item in self.all_detectors if item.cleaning_rule == 'remove;']
        skip_strings = [item.style for item in self.all_detectors if item.cleaning_rule == 'skip;']
        for parsed_element in parsed_elements:
            parsing_string = parsed_element.get('parsing_string')  

            parsing_list = parsing_string.split(';') 
            parsing_list = [parsing for parsing in parsing_list if parsing != '']
            parsing_list = [parsing +';' for parsing in parsing_list]

            # check if ignore
            if any([parsing in remove_strings for parsing in parsing_list]):
                parsed_element.attrib['parsing_type'] = 'remove;'
                #parsed_element.attrib['parsing_log'] += 'clean-parse-ignored;'
            # check if text
            elif any([parsing in skip_strings for parsing in parsing_list]):
                parsed_element.attrib['parsing_type'] = 'skip;'
                #parsed_element.attrib['parsing_log'] += 'clean-parse-text;'
            else:
                parsed_element.attrib['parsing_type'] = parsing_string
                #parsed_element.attrib['parsing_log'] += 'clean-parse-header;'


    # works
    def visualize(self,html):
        # remove style from all descendants so that background color can be set
        for descendant in html.iterdescendants():
            remove_background_color(descendant)

        # find all elements with parsing attribute
        elements = html.xpath('//*[@parsing_type]')
        # get all unique parsing values
        parsing_values = list(set([element.attrib['parsing_type'] for element in elements]))
        color_dict =dict(zip(parsing_values, headers_colors_list[:len(parsing_values)]))
        color_dict.update(self.color_dict)
        # we want header scale, ignore color

        for element in elements:
            # get attribute parsing
            parsing = element.attrib['parsing_type']
            if parsing == '':
                pass
            else:
                color = color_dict[parsing]
                set_background_color(element, color)

        open_tree(html)

    #  code cleanup needed
    def construct_xml_tree(self, html, metadata,add_parsing_id=False):
        root = etree.Element('root')
        document_node = etree.Element('document', title='Document')
        document_node.attrib['parsing_type'] = 'added in tree construction;'
        root.append(document_node)

        levels_dict = {item.style: item.level for item in self.all_detectors if item.level != -1}
        base_headers = [key for key in levels_dict.keys() if levels_dict[key] == 0]


        # ELEMS TO REMEMBER UNTIL NEXT HEADER
        last_section_text = ''
        last_section_title = ''
        last_section_elem = None
        last_section_parsing_type = ''
        last_section_tag = ''
        last_section_parsing_id = None


        # Defining for introduction special case
        last_section_text = ''
        last_section_title = 'Introduction'
        last_section_parsing_type = 'introduction;'
        last_section_tag = 'introduction'
        last_section_parsing_id = '0'


        stack = [document_node]  # added to until header, then modified
        level = None

        flag = True
        remove_elem = None
        for current_event, current_elem in etree.iterwalk(html, events=('start', 'end')): # change to iterwalk to skip certain elements
            if current_event == 'start':

                if remove_elem is not None:
                    continue

                current_parsing_type = current_elem.attrib.get('parsing_type', '')

                if current_parsing_type == 'remove;': # WIP
                    remove_elem = current_elem
                    continue

                if flag: # handle introduction
                    if current_parsing_type in base_headers:
                        flag = False

                        # initialize node using last_section
                        node = etree.Element(last_section_tag, title=last_section_title)
                        node.text = last_section_text
                        node.attrib['parsing_type'] = last_section_parsing_type
                        if add_parsing_id:
                            node.attrib['parsing_id'] = last_section_parsing_id


                        # handle where to append to
                        # append node to document
                        document_node.append(node)
                        # append node to stack
                        stack.append(node)

                        # reset last_section
                        last_section_text = ''
                        last_section_title = clean_title(get_all_text(current_elem))
                        last_section_elem = current_elem
                        last_section_parsing_type = current_parsing_type
                        last_section_parsing_id = str(current_elem.attrib.get('parsing_id'))

                        if current_parsing_type in [key for key in levels_dict.keys()]:
                            level = levels_dict[current_parsing_type]
                            last_section_tag = last_section_parsing_type.replace(';', '')
                        else:
                            level = None
                            last_section_tag = 'company_designated_header'
                    else:
                        last_section_text += get_text(current_elem)
                else:
                    if current_parsing_type == 'skip;':
                        last_section_text += get_text(current_elem)
                    elif current_parsing_type == '':
                        last_section_text += get_text(current_elem)
                    else:
                        # initialize node using last_section
                        node = etree.Element(last_section_tag, title=last_section_title)
                        node.text = last_section_text
                        node.attrib['parsing_type'] = last_section_parsing_type
                        if add_parsing_id:
                            node.attrib['parsing_id'] = last_section_parsing_id

                        # handle where to append to
                        if level is not None: # WIP
                            parent_node = stack[level]
                            stack = stack[:level+1]

                            parent_node.append(node)
                            stack.append(node)

                        else:
                            parsing_type_list = [item.attrib['parsing_type'] for item in stack]
                            if last_section_parsing_type in parsing_type_list:
                                idx = parsing_type_list.index(last_section_parsing_type)
                                
                                if idx == len(stack) - 1:  # If it's the last element
                                    stack[idx-1].append(node)
                                    stack = stack[:idx]  # Remove the last element
                                    stack.append(node)  # Replace with the new node
                                else:
                                    stack = stack[:idx+1]
                                    stack[idx].append(node)
                                    stack.append(node)
                            else:
                                stack[-1].append(node)
                                stack.append(node)

                        # reset last_section
                        last_section_text = ''
                        last_section_title = clean_title(get_all_text(current_elem))
                        last_section_elem = current_elem
                        last_section_parsing_type = current_parsing_type
                        last_section_parsing_id = str(current_elem.attrib.get('parsing_id'))

                        if current_parsing_type in [key for key in levels_dict.keys()]:
                            level = levels_dict[current_parsing_type]
                            last_section_tag = last_section_parsing_type.replace(';', '')
                        else:
                            level = None
                            last_section_tag = 'company_designated_header'
            else:
                if current_elem == remove_elem:
                    remove_elem = None
                    continue

                if current_elem == last_section_elem:
                    last_section_elem = None
                    continue

        # add metadata
        metadata_node = etree.Element('metadata')
        for key in metadata.keys():
            node = etree.Element(key) 
            node.text = str(metadata[key])
            metadata_node.append(node)
        # insert metadata at beginning
        root.insert(0,metadata_node)


        return root

        
# WIP FIX numbers inside pagraphs sometimes parsing as page numbers
class SEC_10K_Parser(HTMLParser):
    def _init(self, **kwargs):
        super()._init(**kwargs)  # Call the parent's _init method

        self.element_detector_group = SEC_10K_ElementGroup()
        self.string_detector_group = SEC_10K_StringDetectorGroup()

        self.color_dict.update({'part;': '#B8860B',
                       'item;': '#BDB76B',
                       'bullet point;': '#FAFAD2',
                          'signatures;': '#B8860B',
                       })


# 10Q and 10K are the same for now
class SEC_10Q_Parser(SEC_10K_Parser):
    pass

class SEC_8K_Parser(HTMLParser):
    def _init(self, **kwargs):
        super()._init(**kwargs)

        self.element_detector_group = SEC_8K_ElementGroup()
        self.string_detector_group = SEC_8K_StringDetectorGroup()
        self.color_dict.update({
                       'item;': '#B8860B',
                       'bullet point;': '#FAFAD2',
                          'signatures;': '#B8860B',
                       })
        
class SEC_S1_Parser(HTMLParser):
    def _init(self, **kwargs):
        super()._init(**kwargs)

        self.element_detector_group = SEC_10K_ElementGroup()
        self.string_detector_group = SEC_S1_StringDetectorGroup()


        self.color_dict.update({
                       'prospectus;': '#B8860B',
                       'item;': '#BDB76B',
                       'signatures;': '#B8860B',
                       'bullet point;': '#FAFAD2',
                       })
        
class SEC_20F_Parser(SEC_10K_Parser):
    pass


import re
# Module to store common regex patterns (move later)
# Module to clean the text

def clean_string(string):
    # Replace zero-width non-joiner
    cleaned_string = string.replace('\u200c', '')
    
    return cleaned_string


def clean_title(text): #  adjust to format items and parts into standardized format, e.g. no periods
    # preprocess
    text = clean_string(text)

    text = text.strip()
    text = re.sub("[^\S \t\n\r\f\v]+",' ',text)

    # remove line breaks
    text = re.sub('\n', ' ', text)

    # replace multiple spaces with single space
    text = re.sub('\s+', ' ', text)


    return text

def clean_string_for_style_detection(text):
    """WIP"""
    # preprocess
    text = clean_string(text)

    chars_to_remove = ['-','_']
    # remove chars_to_remove
    for char in chars_to_remove:
        text = re.sub(char,'',text)

    text = text.strip()
    return text

# common regex patterns
part_pattern = re.compile(r"^part\s+([1234]|iv|i{1,4})(?:$)",re.IGNORECASE)

def is_string_in_middle(parent_text, search_string):
    index = parent_text.find(search_string)
    if index == -1:
        return False  # String not found
    
    if index == 0:
        return False  # String is at the beginning of the text
    
    char_before = parent_text[index - 1]
    return char_before != '\n' and char_before != ''





from lxml import etree
from sec_parsers.parsers_helper import parse_metadata, detect_filing_type, setup_html
from sec_parsers.parsers import SEC_10K_Parser, SEC_10Q_Parser, SEC_8K_Parser, SEC_S1_Parser, SEC_20F_Parser
import csv
import unicodedata


class Filing:
    def __init__(self, html):
        self.metadata = {}
        self._setup_html(html)
        self._parse_metadata()
        self.hierarchy = None # need to implement
        self.xml = None
        self.parser = None # this will be updated by detect_filing_type
        self.filing_type = None
        self._detect_filing_type()
        self._update_parser()


    # keep
    def _setup_html(self,html):
        if type(html).__name__ == 'SEC_Download': # incorporate download metadata
            self.metadata = html.metadata
        self.html = setup_html(html)

    # keep
    def _parse_metadata(self):
        metadata_from_xbrl = parse_metadata(self.html)
        self.metadata.update(metadata_from_xbrl)

    # keep
    def _detect_filing_type(self):
        filing_type = detect_filing_type(self.metadata)
        self.filing_type = filing_type

    def _update_parser(self):
        if self.filing_type is None:
            self._detect_filing_type()
        filing_type = self.filing_type

        if filing_type == '10-K':
            self.parser = SEC_10K_Parser()
        elif filing_type == '10-Q':
            self.parser = SEC_10Q_Parser()
        elif filing_type == '8-K':
            self.parser = SEC_8K_Parser()
        elif filing_type == 'S-1':
            self.parser = SEC_S1_Parser()
        elif filing_type == '20-F':
            self.parser = SEC_20F_Parser()


    # keep
    def set_filing_type(self, filing_type):
        self.filing_type = filing_type
        self._update_parser()

    def parse(self,add_parsing_id=False):
        if self.filing_type is None:
            self._detect_filing_type()

        self.parser.iterative_parse(self.html,add_parsing_id)
        self.parser.clean_parse(self.html)
        # convert to xml
        self._to_xml(add_parsing_id)

    def visualize(self):
        self.parser.visualize(self.html)

    def _to_xml(self,add_parsing_id=False):
        self.xml = self.parser.construct_xml_tree(html=self.html,metadata=self.metadata,add_parsing_id=add_parsing_id)


    # functions to interact with xml

    # Find #
    def find_all_sections_from_title(self,title):
        title = title.strip().lower()
        if self.xml is None:
            self.to_xml()

        # select all nodes with title attribute
        titles = self.xml.xpath(f"//*[@title]")
        # find all nodes with title
        nodes = [node for node in titles if title in node.attrib['title'].lower()]

        return nodes
    
    def find_section_from_title(self,title):
        sections = self.find_all_sections_from_title(title)
        return sections[0] if sections else None

    def find_all_sections_from_text(self, text):
        """Find a node by text."""
        if self.xml is None:
            self.to_xml()

        return self.xml.xpath(f"//*[contains(text(), '{text}')]")
    
    def find_section_from_text(self, text):
        sections = self.find_all_sections_from_text(text)
        return sections[0] if sections else None
    
    # Interact with Node #
    def get_subsections_from_section(self, node):
        """Get all children of a node."""
        return node.getchildren()
    
    def get_nested_subsections_from_section(self, node):
        """Get all children of a node, including nested children."""
        return list(node.iterdescendants())

    # Note, needs refactor, also needs better spacing fix with text.
    def get_text_from_section(self, node, include_title=False):
        """Gets all text from a node, including title string for child nodes."""
        text = ''

        # Add title for child nodes only
        if include_title:
            text += node.attrib.get('title', '') + '\n'

        node_text = node.text
        if node_text is not None:
            text += node.text + '\n'
        
        for child in node:
            # Pass False to indicate this is a child node
            text += self.get_text_from_section(child, include_title=True)
        
        return text
    
    # Interact with tree #

    # TODO: user friendly - e.g. if parsed not called call parse
    def get_tree(self,node=None, level=0):
        if node is None:
            node = self.xml
        tree_string = node.tag
        for child in node:
            tree_string += '\n' + '|-' * level + self.get_tree(child, level + 1)
        return tree_string
    
    def get_title_tree(self,node=None,level=0,attribute='title'):
        if node is None:
            node = self.xml
        elif node.tag == 'metadata':
            return   "Metadata"

            
        tree_atrib = node.attrib.get(attribute,'')
        if 'signatures' in tree_atrib.lower(): #workaround for adding signatures section but not signature parsing
            pass
        else:
            for child in node:
                tree_atrib += '\n' + '|-' * level + self.get_title_tree(child, level + 1,attribute)

        return tree_atrib
    
    # Save to file #
    def save_xml(self, filename, encoding='utf-8'):
        if encoding not in ['utf-8','ascii']:
            raise ValueError('Encoding must be either utf-8 or ascii')
        
        if self.xml is None:
            self.to_xml()

        if encoding == 'ascii':
            self.convert_element_to_ascii(self.xml)

        with open(filename, 'wb') as f:
                f.write(etree.tostring(self.xml,encoding=encoding,xml_declaration=True))

            
    def save_csv(self, filename, encoding='utf-8'):
        if encoding not in ['utf-8','ascii']:
            raise ValueError('Encoding must be either utf-8 or ascii')
        
        if encoding == 'ascii':
            self.convert_element_to_ascii(self.xml)
        
        def get_rows(node,path):
            path = path + "/" + node.attrib.get('title')

            row_list = []

            row = {}
            row['path'] = path
            row['text'] = self.get_text_from_section(node)
            row_list.append(row)
            for child in node:
                row_list.extend(get_rows(child,path))

            return row_list
        
        row_list = []
        for child in self.xml.getchildren():
            row_list.extend(get_rows(child,''))
        
        keys = row_list[0].keys()

        with open(filename, 'w', newline='', encoding='utf-8') as output_file:
            dict_writer = csv.DictWriter(output_file, keys)
            dict_writer.writeheader()
            dict_writer.writerows(row_list)

    # TODO: implement
    def save_dta(self, filename):
        pass

    # encoding
    # Function to convert text to ASCII
    def string_to_ascii(self, string):
        # Normalize to closest ASCII equivalent
        normalized = unicodedata.normalize('NFKD', string)
        # Remove non-ASCII characters
        ascii_string = normalized.encode('ascii', 'ignore').decode('ascii')
        return ascii_string

    # Recursively process all elements and their text
    def convert_element_to_ascii(self,element):
        if element.text:
            element.text = self.string_to_ascii(element.text)
        for child in element:
            self.convert_element_to_ascii(child)
        if element.tail:
            element.tail = self.string_to_ascii(element.tail)
    

        



import requests
import random as rd
import warnings
import re

# WARNING: This code will eventually be removed from SEC Parsers when SEC Downloaders is implemented.

def extract_cik_and_accession(url):
    # Regular expression pattern to match CIK and accession number
    pattern = r"/data/(\d+)/(\d+)/"
    
    # Search for the pattern in the URL
    match = re.search(pattern, url)
    
    if match:
        cik = match.group(1)
        accession_number = match.group(2)
        return cik, accession_number
    else:
        return None, None


# Module to download the SEC filings
#TODO I'm happy with this file so far.

def random_num():
    return rd.randint(1, 1000000)


headers = {
    'User-Agent': f'Sample Company Name {str(random_num)} AdminContact@<sample company domain>.com'
}

def set_headers(name, email):
    """Set the headers for the requests."""
    global headers
    headers['User-Agent'] = f'{name} {email}'

def download_sec_filing(url):
    """Download the SEC filing from the given URL."""
    global headers

    # nudge to set your own headers
    if 'Sample Company Name' in headers['User-Agent']:
        warnings.warn("Warning: Please set your own headers using set_headers(name, email).")

    sec_response = requests.get(url, headers=headers)
    
    if sec_response.status_code != 200:
        print(f"Error {sec_response.status_code}: failed to download {url}")
        return None
    
    # important to encode the text as utf-8 as the lxml parser expects it
    html = sec_response.text.encode("utf-8").decode("utf-8")
    return html



from sec_parsers.css_detectors import BoldCSSDetector, ItalicCSSDetector, UnderlineCSSDetector, HiddenCSSDetector
from sec_parsers.tag_detectors import TableTagDetector, ImageTagDetector, LinkTagDetector, BoldTagDetector, \
    StrongTagDetector, EmphasisTagDetector, ItalicTagDetector, UnderlineTagDetector, TableOfContentsTagDetector


class HeaderElementDetectorGroup:
    def __init__(self):

        self.element_detectors = []
        self.add_element_detector(HiddenCSSDetector(parsing_rule='return',cleaning_rule='remove;'))
        self.add_element_detector(TableTagDetector(parsing_rule='return', cleaning_rule='skip;'))
        self.add_element_detector(ImageTagDetector(parsing_rule='return', cleaning_rule='remove;'))
        self.add_element_detector(LinkTagDetector(parsing_rule='return', cleaning_rule='remove;')) # WIP
        self.add_element_detector(BoldTagDetector(parsing_rule='continue', cleaning_rule='header;'))
        self.add_element_detector(StrongTagDetector(parsing_rule='continue', cleaning_rule='header;'))
        self.add_element_detector(EmphasisTagDetector(parsing_rule='continue', cleaning_rule='header;'))
        self.add_element_detector(ItalicTagDetector(parsing_rule='continue', cleaning_rule='header;'))
        self.add_element_detector(UnderlineTagDetector(parsing_rule='continue', cleaning_rule='header;'))
        self.add_element_detector(BoldCSSDetector(parsing_rule='continue', cleaning_rule='header;'))
        self.add_element_detector(UnderlineCSSDetector(parsing_rule='continue', cleaning_rule='header;'))
        self.add_element_detector(ItalicCSSDetector(parsing_rule='continue', cleaning_rule='header;'))


    def add_element_detector(self, element_detector):
        self.element_detectors.append(element_detector)

    def insert_element_detector(self, element_detector, position):
        self.element_detectors.insert(position, element_detector)

    def detect(self,element):
        for element in self.element_detectors:
            result = element.detect(element)
            if result != '':
                return result
        return ''

class SEC_10K_ElementGroup(HeaderElementDetectorGroup):
    def __init__(self):
        super().__init__()
        # Table of contents detector needs to go before table detector
        self.insert_element_detector(TableOfContentsTagDetector(parsing_rule ='return',cleaning_rule ='skip;'),0)

class SEC_8K_ElementGroup(SEC_10K_ElementGroup):
    pass




class Detector:
    def __init__(self,parsing_rule='continue',cleaning_rule='header;', level = -1):
        self.parsing_rule = parsing_rule # 'return' - marks element as when found to return, 'continue' - marks element as when found to continue
        self.cleaning_rule = cleaning_rule # remove - elem will not be processed in xml, skip- elem will be processed, header - elem will be processed as header
        self.level = level # level of the element in the hierarchy, -1 means no level assigned

class StyleTagDetector(Detector):
    def __init__(self, style=None, **kwargs):
        super().__init__(**kwargs)
        self.style = style # not implemented yet, idea is that e.g. for font weight we can assign 'bold' as group name


from sec_parsers.detectors import StyleTagDetector
from sec_parsers.style_detection import detect_part,detect_item,detect_signatures,detect_page_number,detect_bullet_point,\
    detect_all_caps,detect_note,detect_emphasis_capitalization, detect_empty_string, detect_prospectus

class ProspectusStringDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='prospectus;',**kwargs)

    def detect(self,string):
        if detect_prospectus(string):
            return 'prospectus;'
        else:
            return ''

class PartStringDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='part;',**kwargs)

    def detect(self,string):
        if detect_part(string):
            return 'part;'
        else:
            return ''
        
class ItemStringDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='item;',**kwargs)

    def detect(self,string):
        if detect_item(string):
            return 'item;'
        else:
            return ''
    
class SignaturesStringDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='signatures;',**kwargs)

    def detect(self,string):
        if detect_signatures(string):
            return 'signatures;'
        else:
            return ''
        
class PageNumberStringDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='page number;',**kwargs)

    def detect(self,string): 
        if detect_page_number(string):
            return 'page number;'
        else:
            return ''

class BulletPointStringDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='bullet point;',**kwargs)

    def detect(self,string):
        if detect_bullet_point(string):
            return 'bullet point;'
        else:
            return ''

class AllCapsStringDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='all caps;',**kwargs)

    def detect(self,string):
        if detect_all_caps(string):
            return 'all caps;'
        else:
            return ''

class NoteStringDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='note;',**kwargs)

    def detect(self,string):
        if detect_note(string):
            return 'note;'
        else:
            return ''
              

class EmphasisCapStringDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='emphasis capitalization;',**kwargs)

    def detect(self,string):
        if detect_emphasis_capitalization(string):
            return 'emphasis capitalization;'
        else:
            return ''
        
class EmptyStringDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='empty string;',**kwargs)

    def detect(self,string):
        if detect_empty_string(string):
            return 'empty string;'
        else:
            return ''


# TODO Improve visualiztion

# Important colors (strong contrast, emphasis)
headers_colors_dict = {'part;': '#B8860B',
                       'item;': '#BDB76B',
                       'bullet point;': '#FAFAD2', # LightGoldenRodYellow
                       'table;': '#FAFAD2', # LightGoldenRodYellow
                       'ignore;': '#f2f2f2', # even lighter gray
                          'signatures;': '#cac635',
                       }
headers_colors_list  = [
    "#78CCAB", "#7BCDAD", "#7ECEB9", "#80CFBA", "#83D1BC",
    "#86D2BE", "#88D3C0", "#8BD4C2", "#8ED5C3", "#90D6C5",
    "#93D7C7", "#96D8C9", "#98D9CB", "#9BDACD", "#9EDCCF",
    "#A0DDD0", "#A3DED2", "#A5DFD4", "#A8E0D6", "#ABE1D7",
    "#ADE2D9", "#B0E3DB", "#B3E4DD", "#B5E5DF", "#B8E7E0",
    "#BBE8E2", "#BDE9E4", "#C0EAE6", "#C3EBE8", "#C5ECE9",
    "#C8EDEB", "#CBEEED", "#CDF0EF", "#D0F1F1", "#D3F2F2",
    "#D5F3F4", "#D8F4F5", "#DBF5F7", "#DDF6F8", "#E0F7FA"
]


from sec_parsers.string_detectors import AllCapsStringDetector, EmphasisCapStringDetector, PageNumberStringDetector, BulletPointStringDetector, NoteStringDetector,\
      PartStringDetector, ItemStringDetector, SignaturesStringDetector, NoteStringDetector, EmptyStringDetector, ProspectusStringDetector
from sec_parsers.cleaning import clean_string_for_style_detection

class HeaderStringDetectorGroup:
    def __init__(self):
        self.string_detectors = []
        self.add_string_detector(AllCapsStringDetector())
        self.add_string_detector(EmphasisCapStringDetector())
        self.add_string_detector(EmptyStringDetector(parsing_rule='return',cleaning_rule='remove;'))

    def add_string_detector(self, string_detector):
        self.string_detectors.append(string_detector)

    def insert_string_detectors(self, string_detectors):
        """Inserts string detectors list into first position"""
        self.string_detectors = string_detectors + self.string_detectors

    def detect(self,string):
        string = clean_string_for_style_detection(string)
        for string_detector in self.string_detectors:
            result = string_detector.detect(string)
            if result != '':
                return result
        return ''
    
class SEC_10K_StringDetectorGroup(HeaderStringDetectorGroup):
    def __init__(self):
        super().__init__()
        
        # Add part, item, and signatures detectors at the beginning
        new_detectors = [
            PartStringDetector(parsing_rule='return',level=0,cleaning_rule='header;'),
            ItemStringDetector(parsing_rule='return',level=1,cleaning_rule='header;'),
            SignaturesStringDetector(parsing_rule='return',level=0,cleaning_rule='header;'),
            PageNumberStringDetector(parsing_rule='return',cleaning_rule='remove;'),
            BulletPointStringDetector(parsing_rule='continue',cleaning_rule='skip;')
        ]
        self.insert_string_detectors(new_detectors)
        
class SEC_8K_StringDetectorGroup(HeaderStringDetectorGroup):
    def __init__(self):
        super().__init__()
        
        # Add part, item, and signatures detectors at the beginning
        new_detectors = [
            ItemStringDetector(parsing_rule='return',level=0,cleaning_rule='header;'),
            SignaturesStringDetector(parsing_rule='return',level=0,cleaning_rule='header;'),
            PageNumberStringDetector(parsing_rule='continue',cleaning_rule='remove;'),
            BulletPointStringDetector(parsing_rule='continue',cleaning_rule='skip;')
        ]
        self.insert_string_detectors(new_detectors)

class SEC_S1_StringDetectorGroup(HeaderStringDetectorGroup):
    def __init__(self):
        super().__init__()
        
        # Add part, item, and signatures detectors at the beginning
        new_detectors = [
            ProspectusStringDetector(parsing_rule='return',level=0,cleaning_rule='header;'),
            ItemStringDetector(parsing_rule='return',level=0,cleaning_rule='header;'),
            SignaturesStringDetector(parsing_rule='return',level=0,cleaning_rule='header;'),
            PageNumberStringDetector(parsing_rule='continue',cleaning_rule='remove;'),
            BulletPointStringDetector(parsing_rule='continue',cleaning_rule='skip;')
        ]
        self.insert_string_detectors(new_detectors)
        


from lxml import etree
from sec_parsers.style_detection import get_all_text
import warnings

# WIP: convert special checked boxes into True / False, etc
def parse_metadata(html):
    """WIP"""
    metadata_dict = {}
    # select all elements with attribute name and first part of attribute is 'dei'
    elements = html.xpath('//*[@name]')
    elements = [element for element in elements if element.attrib['name'].lower().startswith('dei')]
    for element in elements:
        name = element.attrib['name']
        name = name.replace('dei:','')
        metadata_dict[name] = get_all_text(element)
    return metadata_dict

def detect_filing_type(metadata):
    """WIP"""
    if 'DocumentType' in metadata.keys():
        return metadata['DocumentType']
    else:
        warnings.warn('DocumentType not found in metadata. Filing type set to 10K. If this is not correct, please set the filing type manually.')
        return '10-K'

def setup_html(html):

        # Find the start of the HTML content. This is necessary because the HTML content is not always at the beginning of the file.
        body_start = html.find('<BODY')
        if body_start == -1:
            body_start = html.find('<body')

        if body_start != -1:
            html = html[body_start:]

        parser = etree.HTMLParser(encoding='utf-8',remove_comments=True)
        html = etree.fromstring(html, parser)

        return html


from sec_parsers.detectors import StyleTagDetector
from sec_parsers.style_detection import detect_link,detect_bold_from_html,detect_strong_from_html,detect_emphasis_from_html,\
    detect_italic_from_html,detect_underline_from_html,detect_table,detect_table_of_contents,detect_image

class LinkTagDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='link;', **kwargs)

    def detect(self,element):
        if detect_link(element):
            return 'link;'
        else:
            return ''
        
class BoldTagDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='bold;', **kwargs)

    def detect(self,element):
        if detect_bold_from_html(element):
            return 'bold;'
        else:
            return ''

class StrongTagDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style ='strong;', **kwargs)

    def detect(self,element):
        if detect_strong_from_html(element):
            return 'strong;'
        else:
            return ''
        
class EmphasisTagDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='emphasis;',**kwargs)

    def detect(self,element):
        if detect_emphasis_from_html(element):
            return 'emphasis;'
        else:
            return ''
        
class ItalicTagDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='italic;',**kwargs)

    def detect(self,element):
        if detect_italic_from_html(element):
            return 'italic;'
        else:
            return ''

class UnderlineTagDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='underline;',**kwargs)

    def detect(self,element):
        if detect_underline_from_html(element):
            return 'underline;'
        else:
            return ''
        

        
class TableTagDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='table;',**kwargs)

    def detect(self,element):
        if detect_table(element):
            return 'table;'
        
        return ''
        
class TableOfContentsTagDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='table of contents;',**kwargs)

    def detect(self,element):
        if detect_table(element):
            if detect_table_of_contents(element) == "toc":
                return 'table of contents;'
            
        return ''
        
class ImageTagDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='image;', **kwargs)

    def detect(self,element):
        if detect_image(element):
            return 'image;'
        else:
            return ''
        


from sec_parsers.detectors import StyleTagDetector
from sec_parsers.style_detection import detect_hidden_element, detect_bold_from_css, detect_underline_from_css, detect_italic_from_css

class HiddenCSSDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='hidden;',**kwargs)

    def detect(self,element):
        if detect_hidden_element(element):
            return 'hidden;'
        else:
            return ''
        
class BoldCSSDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='bold;',**kwargs)

    def detect(self,element):
        if detect_bold_from_css(element):
            return 'bold;'
        else:
            return ''
        
class UnderlineCSSDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='underline;',**kwargs)

    def detect(self,element):
        if detect_underline_from_css(element):
            return 'underline;'
        else:
            return ''
        
class ItalicCSSDetector(StyleTagDetector):
    def __init__(self, **kwargs):
        super().__init__(style='italic;',**kwargs)

    def detect(self,element):
        if detect_italic_from_css(element):
            return 'italic;'
        else:
            return ''


from setuptools import setup, find_packages

from pathlib import Path
long_description = Path("../readme.md").read_text()

setup(
    name="sec_parsers",
    author="John Friedman",
    version="0.549",
    description = "A package to parse SEC filings",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    url="https://github.com/john-friedman/SEC-Parsers",
    install_requires=[
        'lxml', 
        'requests',
    ],
    extras_require={
        'downloaders': ['sec_downloaders'],
        'visualizers': ['sec_visualizers'],
        'all': ['sec_downloaders', 'sec_visualizers'],
    },
)


import re
from sec_parsers.xml_helper import get_all_text
from sec_parsers.cleaning import clean_string_for_style_detection, part_pattern

#TODO happy with this file so far

# simple for now
def is_paragraph(text):
    periods = text.count('.')
    commas = text.count(',')
    if periods > 0:
        return True
    
    return False

# TODO: improve this
def detect_bullet_point(string):
    """e.g. •"""
    if any(char in string.strip() for char in ['•','●','●','●','●','•','·','◦']):
        return True
    return False

def detect_signatures(string):
    """e.g. Signatures"""
    match = re.search(r"^SIGNATURES$",string, re.IGNORECASE)
    if match:
        return True
    
    return False

def detect_item(string):
    """e.g. Item 1A. Risk Factors"""
    match = re.search(r"^Item\s+\d+[A-Z]{0,}",string, re.IGNORECASE)
    if match:
        return True
    return False

def detect_emphasis_capitalization(string): # WIP
    """Seen in amazon's 2024 10k e.g. We Have Foreign Exchange Risk"""
    # Check for None, None., Omitted., Not Applicable.
    if string.lower() in [item.lower() for item in ['None','None.','Omitted.', 'Not Applicable.']]:
        return False
    
    # if one word return false
    if len(string.split()) == 1:
        return False
    
    # if all caps
    if string.isupper():
        return False
    
    # Check if string is all digits
    if string.isdigit():
        return False
    
    words = string.split()
    if not words:
        return False

    has_capitalized_word = False
    for word in words:
        # Skip prepositions, articles, and conjunctions
        if word.lower() in ["of",'to','a','and','by','in','the','or','on','for','as','with','that','but','not','so','yet','an','at','off','per','up','via']:
            continue
        
        # Check if the word contains any alphabetic characters
        if any(char.isalpha() for char in word):
            if word[0].isupper():
                has_capitalized_word = True
            else:
                return False
        # If the word is just digits or punctuation, ignore it

    return has_capitalized_word
    
def detect_prospectus(string):
    """e.g. PROSPECTUS SUMMARY"""
    match = re.search(r"^PROSPECTUS SUMMARY$",string, re.IGNORECASE)
    if match:
        return True
    return False    

def detect_part(string):
    """e.g. Part I"""
    match = part_pattern.match(string.lower().strip())
    if match:
        return True
    return False
    
# deprecated for now
def level_detection(string):
    """e.g. amazon Level 1"""
    match = re.search(r"^Level\s+\d+",string)
    if match:
        return True
    return False
    

def detect_note(string):
    """e.g. Note 1"""
    match = re.search(r"^Note\s+\d+",string, re.IGNORECASE)
    if match:
        return True
    return False

def detect_all_caps(string):
    """e.g. FORM 10-K SUMMARY"""
    if string.isupper():
        return True
    # may cause issue
    elif string.isdigit():
        return True
    return False

def detect_page_number(string):
    if re.search(r"^\d+$",string, re.IGNORECASE):
        return True
    elif re.search(r'F-\d+$',string, re.IGNORECASE):
        return True
    else:
        return False

def detect_style_from_string(string):
    
    # PREPROCESSING WIP
    string = clean_string_for_style_detection(string)
    # ORDER MATTERS
    if detect_part(string):
        return 'part;'
    elif detect_item(string):
        return 'item;'
    elif detect_signatures(string):
        return 'signatures;'
    elif detect_page_number(string):
        return 'page number;'
    elif detect_bullet_point(string):
        return 'bullet point;'
    elif detect_all_caps(string):
        return 'all caps;'
    elif detect_emphasis_capitalization(string):
        return 'emphasis;'
    elif detect_note(string):
        return 'note;'
    else:
        return ''
    

def detect_hidden_element(element):
    if element.get('style'):
        if 'display:none' in re.sub(' ','',element.get('style')):
            return True
    return False

def detect_link(node):
    """Detects if a node is a link."""
    if node.tag == 'a':
        return True
    return False

def detect_bold_from_html(element):
    """Detects bold from html tags"""
    tag_list = ['b']
    if element.tag in tag_list:
        return True
    return False

def detect_strong_from_html(element):
    """Detects strong from html tags"""
    tag_list = ['strong']
    if element.tag in tag_list:
        return True
    return False

def detect_emphasis_from_html(element):
    """Detects emphasis from html tags"""
    tag_list = ['em']
    if element.tag in tag_list:
        return True
    return False

def detect_italic_from_html(element):
    """Detects italic from html tags"""
    tag_list = ['i']
    if element.tag in tag_list:
        return True
    return False

def detect_underline_from_html(element):
    """Detects underline from html tags"""
    tag_list = ['u']
    if element.tag in tag_list:
        return True
    return False

# deprecated
def detect_from_html(element):
    element_style = ''
    ancestors = list(element.iterancestors())
    # include self
    ancestors = ancestors + [element]
    # detect bold from ancestors
    for ancestor in ancestors:
        tag_list = ['b','strong','em','i','u']
        if ancestor.tag in tag_list:
            element_style += f'{ancestor.tag}-tag;'
        
    return element_style


# WIP, may need modification for different font weights
def detect_bold_from_css(element):
    """Detects bold from css"""
    if element.get('style'):
        if 'font-weight:bold' in element.get('style'):
            return True
        # change to be any font weight greater than 400
        elif 'font-weight:700' in element.get('style'):
            return True
    return False



def detect_underline_from_css(element):
    """Detects underline from css"""
    if element.get('style'):
        if 'text-decoration:underline' in element.get('style'):
            return True
    return False

def detect_italic_from_css(element):
    """Detects italic from css"""
    if element.get('style'):
        if 'font-style:italic' in element.get('style'):
            return True
    return False


def detect_style_from_element(element):
    # check it or descendants have text
    text = get_all_text(element).strip() # WIP DOES THIS MATTER? WE WILL SEE
    if len(text) == 0:
        return ''

    
    style = ''
    style += detect_bold_from_css(element)
    style += detect_from_html(element)
    style += detect_underline_from_css(element)
    style += detect_italic_from_css(element)
    style += detect_link(element)
    return style

def is_descendant_of_table(element):
    return bool(element.xpath("./ancestor::table"))

# needs work 
def detect_table(table):
    """Detects if table or header disguised as a table"""
    if table.tag != 'table':
        return False
    
    # this doesn't work
    #tr_list = table.xpath('//tr')

    tr_list = table.findall('tr')
    if len(tr_list) > 3:
        return True
    text = get_all_text(table)

    if detect_bullet_point(text):
        return False
    number_count = len(re.findall(r'\d', text))
    if number_count > 5:
        return True
    
    char_count = len(text)
    if char_count > 400:
        return True
    
    return False


def detect_table_of_contents(element):
    toc_type = 'not-toc'
    """Detects if a table is likely to be a table of contents."""

    # toc - needs seperate parser if no links
    num_items = len(re.findall('Item(\s+|$|\n)', get_all_text(element), re.IGNORECASE))
    if num_items > 9:
        toc_type = 'toc'

    # linked toc
    # links = element.find_all('a')
    # if len(links) > 5:
    #     toc_type = 'toc-links'

    return toc_type

def detect_toc_link(node):
    """Detects if a node is a table of contents link."""
    if node.tag == 'a':
        text = get_all_text(node)
        if text.lower() in ['table of contents','toc']:
            return True
    return False

def detect_image(node):
    """Detects if a node is an image."""
    if node.tag == 'img':
        return True
    return False

def detect_empty_string(string):
    return bool(re.match(r'^\s*$', string))

